import random
import pytest
from ex1 import concat, factorial2, length2, reversed, drop, factorial, get_item, hd, head, last, length, nth, reversed2, take

# random.seed(1)

# list of identifiers for the generated test cases (default number is 10)
test_ids = ["test_" + str(i) for i in range(10)]


def test_length() -> None:
    expected = 4
    actual = length([10, 21, 32, 43])
    assert expected == actual


@pytest.mark.parametrize(("l", "expected"), [
    ([], 0),
    ([1, 2, 3], 3),
    ([1, 2, 3, 4, 5], 5),
])
def test_length_param(l: list[int], expected: int) -> None:
    result = length(l)
    assert result == expected


@pytest.mark.parametrize(("l"), [
    ([]),
])
def test_length_property1(l: list[int]) -> None:
    result = length(l)
    assert l == []
    assert result == 0


@pytest.mark.parametrize(("l"), [
    ([1, 2, 3]),
    ([1, 2, 3, 4, 5]),
])
def test_length_property2(l: list[int]) -> None:
    result = length(l)
    assert l != []
    [_, *t] = l
    assert result == 1 + length(t)


@pytest.mark.parametrize(("l"), [
    ([]),
])
def test_length2_property1(l: list[int]) -> None:
    result = length2(l)
    assert l == []
    assert result == 0


@pytest.mark.parametrize(("l"), [
    ([1, 2, 3]),
    ([1, 2, 3, 4, 5]),
])
def test_length2_property2(l: list[int]) -> None:
    result = length2(l)
    assert l != []
    [_, *t] = l
    assert result == 1 + length2(t)


@pytest.mark.parametrize(("l"), [
    ([]),
    ([1, 2, 3]),
    ([1, 2, 3, 4, 5]),
])
def test_length2_property3(l: list[int]) -> None:
    result = length2(l)
    assert result == length(l)


def test_hd() -> None:
    expected = 10
    actual = hd([10, 21, 32, 43])
    assert expected == actual


@pytest.mark.parametrize(("l"), [
    ([]),
])
def test_hd_property1(l: list[int]) -> None:
    with pytest.raises(ValueError):
        hd(l)


@pytest.mark.parametrize(("l"), [
    ([1, 2, 3]),
    ([1, 2, 3, 4, 5]),
])
def test_hd_property2(l: list[int]) -> None:
    assert l != []
    result = hd(l)
    [h, *_] = l
    assert result == h


@pytest.mark.parametrize(("l"), [
    ([]),
    ([1, 2, 3]),
    ([1, 2, 3, 4, 5]),
])
def test_hd_property3(l: list[int]) -> None:
    match l:
        case [h, *_]:
            result = hd(l)
            assert result == h
        case _:
            with pytest.raises(ValueError):
                hd(l)


def test_head() -> None:
    expected = 10
    actual = head([10, 21, 32, 43])
    assert expected == actual


@pytest.mark.parametrize(("l"), [
    ([]),
])
def test_head_property1(l: list[int]) -> None:
    result = head(l)
    assert result == None


@pytest.mark.parametrize(("l"), [
    ([1, 2, 3]),
    ([1, 2, 3, 4, 5]),
])
def test_head_property2(l: list[int]) -> None:
    assert l != []
    result = head(l)
    [h, *_] = l
    assert result == h


@pytest.mark.parametrize(("l"), [
    ([]),
    ([1, 2, 3]),
    ([1, 2, 3, 4, 5]),
])
def test_head_property3(l: list[int]) -> None:
    match l:
        case [h, *_]:
            result = head(l)
            assert result == h
        case _:
            result = head(l)
            assert result == None


def test_get_item() -> None:
    expected: tuple[int, list[int]] = (10, [21, 32, 43])
    actual = get_item([10, 21, 32, 43])
    assert actual is not None
    assert expected == actual

def test_get_item2() -> None:
    actual = get_item([])
    assert actual is None

def test_last() -> None:
    expected = 43
    actual = last([10, 21, 32, 43])
    assert expected == actual

def test_last2() -> None:
    with pytest.raises(ValueError):
        last([])

def test_nth() -> None:
    expected = 32
    actual = nth([10, 21, 32, 43], 2)
    assert expected == actual


def test_take() -> None:
    expected = [10, 21]
    actual = take([10, 21, 32, 43], 2)
    assert expected == actual


def test_drop() -> None:
    expected = [32, 43]
    actual = drop([10, 21, 32, 43], 2)
    assert expected == actual


def test_reversed() -> None:
    expected = [43, 32, 21, 10]
    actual = reversed([10, 21, 32, 43])
    assert expected == actual


@pytest.mark.parametrize(("l"), [
    ([]),
    ([1, 2, 3]),
    ([1, 2, 3, 4, 5]),
])
def test_reversed_property(l: list[int]) -> None:
    match l:
        case [h, *t]:
            result = reversed(l)
            assert result == reversed(t) + [h]
        case _:
            result = reversed(l)
            assert result == []


@pytest.mark.parametrize(("l"), [
    ([]),
    ([1, 2, 3]),
    ([1, 2, 3, 4, 5]),
])
def test_reversed2_property(l: list[int]) -> None:
    match l:
        case [h, *t]:
            result = reversed2(l)
            assert result == reversed2(t) + [h]
        case _:
            result = reversed2(l)
            assert result == []


@pytest.mark.parametrize(("l"), [
    ([]),
    ([1, 2, 3]),
    ([1, 2, 3, 4, 5]),
])
def test_reversed2_property2(l: list[int]) -> None:
    result = reversed2(l)
    assert result == reversed(l)


def test_concat() -> None:
    expected = [1, 2, 3, 4, 5, 6]
    actual = concat([[1, 2], [], [3, 4, 5], [6]])
    assert expected == actual
